#include <algorithm>
#include <iostream>
#include <memory>

class X {
public:
	X() { std::cout << "X constructed\n"; }
	~X() { std::cout << "X destroyed\n"; }
	void nonConst() { std::cout << "non-const method called\n"; }
};

void foo1(std::unique_ptr<X> ptr) {
	// ptr is now owned by foo1
	std::cout << "in foo1\n";
	// the memory is deallocated at the end of scope here
}

void foo2(std::unique_ptr<X>& ptr) {
	std::cout << "in foo2, before reset\n";
	// foo2 can do anything with ptr, including reset and =
	ptr.reset();
	std::cout << "in foo2, after reset\n";
}

void foo3(const std::unique_ptr<X>& ptr) {
	// ve cannot change ptr (via reset or =)
	// but we can change the pointed-at object
	ptr->nonConst();
}

void foo4(const std::unique_ptr<const X>& ptr) {
	// we cannot call this
}

void foo5(X* ptr) {
	// we can change the pointed-at object
	ptr->nonConst();
}

void foo6(const X* ptr) {
	// we cannot change the pointed-at object
	// ptr->nonConst();
}

void foo7(X& ref) {
	// we do not have to test for nullptr
	// we can change the referenced object
	ref.nonConst();
}

void foo8(const X& ref) {
	// we do not have to test for nullptr
	// we cannot change the referenced object
	// ref.nonConst();
}

// note: all the above functions except for foo7 and foo8 should check
// whether their parameter is nullptr
// in the case of foo7 a foo8 it is the caller's responsibility

int main() {
	auto ptr = std::make_unique<X>();

	//foo1(ptr); // does not work, we cannot copy ownership
	foo1(move(ptr)); // we move the ownership
	// ptr does not own anything anymore
	std::cout << "---\n";

	ptr = std::make_unique<X>();
	foo2(ptr);
	// does ptr own anything? how can we know?

	std::cout << "---\n";

	ptr = std::make_unique<X>();
	// kind-of OK, if we want to allow object modification
	// (it is needlessly complicated, however)
	foo3(ptr);

	std::cout << "---\n";

	//foo4(ptr); // does not work; why?

	foo5(ptr.get());

	std::cout << "---\n";

	foo6(ptr.get());

	std::cout << "---\n";

	foo7(*ptr);

	std::cout << "---\n";

	foo8(*ptr);

	std::cout << "---\n";
}
